"""Steps for testing the performance of queries."""
from behave import then


@then('the step `{step}` execution should only take "{max_seconds}" seconds')
def check_execution_time(context, step, max_seconds):
    """Check that the step only takes the specified time."""
    context.test.assertTrue(context.step_execution_time[step] <= float(max_seconds))



"""Response data steps."""
from behave import then


@then("values exist in the response")
def response_values_exist(context):
    """Check that values exist in the response."""
    context.test.assertTrue(context.table)
    context.test.assertTrue(context.table[0].as_dict())
    for row in context.table:
        for key, value in row.as_dict().items():
            response_value = context.response.data.get(key)
            if isinstance(response_value, list):
                context.test.assertIn(value, response_value)
            else:
                context.test.assertEqual(
                    response_value, value, f"Key: {key}, Value: {value}"
                )


@then('values exist in "{response_key}" in the response')
def response_values_exist_at_key(context, response_key):
    """Check that values exist in the response."""
    context.test.assertTrue(context.table)
    context.test.assertTrue(context.table[0].as_dict())
    data = context.response.data.get(response_key, {})
    for row in context.table:
        for key, value in row.as_dict().items():
            if isinstance(data, list):
                found = False
                for item in data:
                    response_value = item.get(key)
                    if response_value == value:
                        found = True
                        break
                context.test.assertTrue(found)
            else:
                response_value = data.get(key)
                response_type = type(response_value)
                if not isinstance(value, response_type):
                    if response_type == bool:
                        value = value == "True"
                    else:
                        value = type(response_value)(value)
                context.test.assertEqual(
                    response_value, value, f"Key: {key}, Value: {value}"
                )



"""Steps for testing authorization."""
from behave import then
from django.contrib.auth.models import Group


@then('I should have the role "{role_name}"')
def has_role(context, role_name):
    """Check that the user has the role.

    Args:
        context (behave.runner.Context): The test context.
        role_name (str): The name of the role.
    """
    context.test.assertTrue(getattr(context, "user", None) is not None)
    context.test.assertTrue(context.user.groups.filter(name=role_name).exists())


@then('I should have the group "{group_name}"')
def has_group(context, group_name):
    """Check that the user has the group.

    Args:
        context (behave.runner.Context): The test context.
        group_name (str): The name of the group.
    """
    context.execute_steps(f'Then I should have the role "{group_name}"')


@then('I should not have the role "{role_name}"')
def does_not_have_role(context, role_name):
    """Check that the user does not have the role.

    Args:
        context (behave.runner.Context): The test context.
        role_name (str): The name of the role.
    """
    context.test.assertTrue(getattr(context, "user", None) is not None)
    context.test.assertFalse(context.user.groups.filter(name=role_name).exists())


@then('I should not have the group "{group_name}"')
def does_not_have_group(context, group_name):
    """Check that the user does not have the group.

    Args:
        context (behave.runner.Context): The test context.
        group_name (str): The name of the group.
    """
    context.execute_steps(f'Then I should not have the role "{group_name}"')


@then('I should have the permission "{permission_name}" for model "{model_name}"')
def has_permission(context, permission_name, model_name):
    """Check that the user has the role.

    Args:
        context (behave.runner.Context): The test context.
        permission_name (str): The name of the role.
        model_name (str): The name of the model.
    """
    context.test.assertTrue(getattr(context, "user", None) is not None)
    context.execute_steps(f'Given a "{model_name}" model is available')
    app_label = context.models[  # pylint: disable=protected-access
        model_name
    ]._meta.app_label
    context.test.assertTrue(context.user.has_perm(f"{app_label}.{permission_name}"))


@then('I should not have the permission "{permission_name}" for model "{model_name}"')
def does_not_have_permission(context, permission_name, model_name):
    """Check that the user does not have the role.

    Args:
        context (behave.runner.Context): The test context.
        permission_name (str): The name of the role.
        model_name (str): The name of the model.
    """
    context.test.assertTrue(getattr(context, "user", None) is not None)
    context.execute_steps(f'Given a "{model_name}" model is available')
    app_label = context.models[  # pylint: disable=protected-access
        model_name
    ]._meta.app_label
    context.test.assertFalse(context.user.has_perm(f"{app_label}.{permission_name}"))


@then(
    'the role "{dest_role}" should have the permissions from the "{source_role}" role'
)
def inherit_permissions_from_role(context, dest_role, source_role):
    """Check the role has the permissions from the other role.

    Args:
        context (behave.runner.Context): The test context.
        dest_role (str): The role to give the permissions to.
        source_role (str): The role to inherit the permissions from.
    """
    source_group = Group.objects.filter(name=source_role).first()
    context.test.assertTrue(source_group is not None)
    dest_group = Group.objects.filter(name=dest_role).first()
    context.test.assertTrue(dest_group is not None)
    permissions = source_group.permissions.all()
    for permission in permissions:
        context.test.assertTrue(
            dest_group.permissions.filter(pk=permission.pk).exists()
        )



"""Authentication steps."""
from behave import given
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType


@given('I am assigned to the group "{group_name}"')
def assign_user_to_group(context, group_name):
    """Assigns a user to a group.

    Args:
        context: behave.runner.Context
        group_name: str
    """
    if not getattr(context, "user", None):
        context.execute_steps("Given I am a registered user")
    if not getattr(context, "models", None):
        context.execute_steps('Given a "Group" model is available')
    context.execute_steps(
        f"""Given a "Group" with values
                          | name |
                          | {group_name} |"""
    )
    model = context.models.get("Group")
    context.user.groups.add(model.objects.get(name=group_name))


@given('I am assigned to the role "{role_name}"')
def assign_user_to_role(context, role_name):
    """Checks if a user is assigned to a role.

    Args:
        context: behave.runner.Context
        role_name: str
    """
    context.execute_steps(f'Given I am assigned to the group "{role_name}"')


@given(
    'The group "{group_name}" has the permission "{permission_name}"'
    ' with codename "{codename}" for model "{model_name}"'
)
def give_group_permission(context, group_name, permission_name, codename, model_name):
    """Gives a group a permission for a model.

    Args:
        context: behave.runner.Context
        group_name: str
        permission_name: str
        codename: str
        model_name: str
    """
    context.execute_steps('Given a "Group" model is available')
    context.execute_steps(
        f"""Given a "Group" with values
                          | name |
                          | {group_name} |"""
    )
    model = context.models.get("Group")
    group = model.objects.get(name=group_name)
    context.execute_steps(
        f'Given the permission "{permission_name}" with codename "{codename}" '
        f'for model "{model_name}" exists'
    )
    model = context.models.get(model_name)
    content_type = ContentType.objects.get_for_model(model)
    permission = Permission.objects.get(codename=codename, content_type=content_type)
    group.permissions.add(permission)


@given(
    'the permission "{permission_name}" with codename "{codename}" '
    'for model "{model_name}" exists'
)
def create_permission_for_model(context, permission_name, codename, model_name):
    """Creates a permission for a model.

    Args:
        context: behave.runner.Context
        permission_name: str
        codename: str
        model_name: str
    """
    context.execute_steps(f'Given a "{model_name}" model is available')
    model = context.models.get(model_name)
    content_type = ContentType.objects.get_for_model(model)
    if not Permission.objects.filter(
        codename=codename, content_type=content_type
    ).exists():
        Permission.objects.create(
            codename=codename,
            name=permission_name,
            content_type=content_type,
        )


@given('the role "{dest_role}" inherits permissions from the "{source_role}" role')
def given_role_inherit_permissions_from_role(context, dest_role, source_role):
    """Give the role the permissions from the other role.

    Args:
        context (behave.runner.Context): The test context.
        dest_role (str): The role to give the permissions to.
        source_role (str): The role to inherit the permissions from.
    """
    source_group = Group.objects.filter(name=source_role).first()
    context.test.assertTrue(source_group is not None)
    dest_group = Group.objects.filter(name=dest_role).first()
    context.test.assertTrue(dest_group is not None)
    permissions = source_group.permissions.all()
    dest_group.permissions.add(*permissions)



"""Execute request steps."""
from behave import step, when


@step("api client is available")
def api_client_is_available(context):
    """API client is available."""
    try:
        from rest_framework.test import (  # pylint: disable=import-outside-toplevel
            APIClient,
        )

        context.test.api_client = APIClient()
    except ImportError as exc:
        raise ImportError("Django REST Framework is not installed.") from exc


@when('I make a "{method}" request to "{url}"')
def make_request(context, method, url):
    """Make a request."""
    context.response = getattr(context.test.client, method.lower())(
        url, headers=context.headers if hasattr(context, "headers") else None
    )


@when('I make a "{method}" request to "{url}" with data')
def make_request_with_data(context, method, url):
    """Make a request with data."""
    context.response = getattr(context.test.client, method.lower())(
        url,
        data=context.request_data,
        headers=context.headers if hasattr(context, "headers") else None,
    )


@when('I make an API "{method}" request to "{url}"')
def make_api_request(context, method, url):
    """Make an API request."""
    context.execute_steps("Given api client is available")
    context.response = getattr(context.test.api_client, method.lower())(
        url,
        format="json",
        headers=context.headers if hasattr(context, "headers") else None,
    )


@when('I make an API "{method}" request to "{url}" with data')
def make_api_request_with_data(context, method, url):
    """Make an API request with data."""
    context.execute_steps("Given api client is available")
    context.response = getattr(context.test.api_client, method.lower())(
        url,
        data=context.request_data,
        format="json",
        headers=context.headers if hasattr(context, "headers") else None,
    )



"""Steps for testing the performance of queries."""
import time
from collections import defaultdict

from behave import when


@when('I execute the step `{step}` it should make "{query_count}" queries')
def check_query_count(context, step, query_count):
    """Check that the step only makes the expected number of queries."""
    with context.test.assertNumQueries(int(query_count)):
        context.execute_steps(step)


@when("I time the step `{step}`")
def check_execution_time(context, step):
    """Time the step."""
    before = time.time()
    context.execute_steps(step)
    after = time.time()
    if not hasattr(context, "step_execution_time"):
        context.step_execution_time = defaultdict(lambda: 0.0)
    context.step_execution_time[step] = after - before



"""Authentication steps."""
import base64

from behave import given
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser

HTTP_HEADER_ENCODING = "iso-8859-1"

User = get_user_model()


def basic_auth_header(username: str, password: str) -> str:
    """Return a dict with the basic auth header.

    Args:
        username (str): The username.
        password (str): The password.
    """
    credentials = f"{username}:{password}"
    base64_credentials = base64.b64encode(
        credentials.encode(HTTP_HEADER_ENCODING)
    ).decode(HTTP_HEADER_ENCODING)
    return f"Basic {base64_credentials}"


@given('the user has a password of "{password}"')
def user_with_password(context, password):
    """Setup password.

    Args:
        context (behave.runner.Context): The test context.
        password (str): The password to use.
    """
    context.user_password = password


@given('the user has a username of "{username}"')
def user_with_username(context, username):
    """Setup username.

    Args:
        context (behave.runner.Context): The test context.
        username (str): The username to use.
    """
    context.user_username = username


@given('the user has a first name of "{first_name}"')
def user_with_firstname(context, first_name):
    """Setup first_name.

    Args:
        context (behave.runner.Context): The test context.
        first_name (str): The first name to use.
    """
    context.user_first_name = first_name


@given('the user has a last name of "{last_name}"')
def user_with_lastname(context, last_name):
    """Setup last_name.

    Args:
        context (behave.runner.Context): The test context.
        last_name (str): The username to use.
    """
    context.user_last_name = last_name


@given("the user is staff")
def user_with_staff(context):
    """Setup staff user.

    Args:
        context (behave.runner.Context): The test context.
    """
    context.user_is_staff = True


@given("the user is a superuser")
def user_with_superuser(context):
    """Setup superuser.

    Args:
        context (behave.runner.Context): The test context.
    """
    context.user_is_superuser = True


@given("the user is active")
def user_is_active(context):
    """Setup superuser.

    Args:
        context (behave.runner.Context): The test context.
    """
    context.user_is_active = True


@given("the user is not active")
def user_is_not_active(context):
    """Setup superuser.

    Args:
        context (behave.runner.Context): The test context.
    """
    context.user_is_active = False


@given("I am a user")
def is_user(context):
    """Setup superuser.

    Args:
        context (behave.runner.Context): The test context.
    """
    user = User.objects.create_user(
        username=getattr(context, "user_username", "test"),
        password=getattr(context, "user_password", "test"),
        is_active=getattr(context, "user_is_active", True),
        is_staff=getattr(context, "user_is_staff", False),
        is_superuser=getattr(context, "user_is_superuser", False),
        first_name=getattr(context, "user_first_name", "Test"),
        last_name=getattr(context, "user_last_name", "User"),
    )
    context.user = user
    context.test.client.force_login(user)


@given("I am a superuser")
def is_superuser(context):
    """Setup superuser.

    Args:
        context (behave.runner.Context): The test context.
    """
    user = User.objects.create_user(
        username=getattr(context, "user_username", "test_superuser"),
        password=getattr(context, "user_password", "test"),
        is_active=getattr(context, "user_is_active", True),
        is_staff=getattr(context, "user_is_staff", False),
        is_superuser=True,
        first_name=getattr(context, "user_first_name", "Test"),
        last_name=getattr(context, "user_last_name", "Superuser"),
    )
    context.user = user
    context.test.client.force_login(user)


@given("I am a registered user")
def setup_registered_user(context):
    """Setup registered user.

    Args:
        context (behave.runner.Context): The test context.
    """
    print(context)
    for key, value in context.__dict__.items():
        print(key, value)
    user = User.objects.create_user(
        username=getattr(context, "user_username", "test_user"),
        password=getattr(context, "user_password", "test"),
        is_active=getattr(context, "user_is_active", True),
        is_staff=getattr(context, "user_is_staff", False),
        is_superuser=getattr(context, "user_is_superuser", False),
        first_name=getattr(context, "user_first_name", "Test"),
        last_name=getattr(context, "user_last_name", "User"),
    )
    context.user = user
    context.test.client.force_login(user)


@given("I am a staff user")
def setup_staff_user(context):
    """Setup staff user.

    Args:
        context (behave.runner.Context): The test context.
    """
    user = User.objects.create_user(
        username=getattr(context, "user_username", "test_staff_user"),
        password=getattr(context, "user_password", "test"),
        is_active=getattr(context, "user_is_active", True),
        is_staff=True,
        is_superuser=getattr(context, "user_is_superuser", False),
        first_name=getattr(context, "user_first_name", "Test"),
        last_name=getattr(context, "user_last_name", "StaffUser"),
    )
    context.user = user
    context.test.client.force_login(user)


@given("I am an anonymous user")
def setup_anon_user(context):
    """Setup anonymous user.

    Args:
        context (behave.runner.Context): The test context.
    """
    context.user = AnonymousUser()
    context.credentials = basic_auth_header("test", "bad_password")



"""Fixtures steps."""
from copy import copy

from behave import given
from behave.runner import Context


@given('these fixtures are used "{fixture_names}"')
def step_impl(context: Context, fixture_names: str):
    """Load fixtures.

    This step is used to load fixtures in the context of a test.

    Args:
        context (Context): The behave context.
        fixture_names (str): A comma separated list of fixture names.
    """
    fixture_names = (
        fixture_names.split(",") if "," in fixture_names else [fixture_names]
    )
    context.fixtures = fixture_names
    context.test.fixtures = copy(context.fixtures)
    # pylint: disable=import-outside-toplevel
    from django.core.management import call_command

    # pylint: disable-next=protected-access
    for db_name in context.test._databases_names(include_mirrors=False):
        try:
            call_command(
                "loaddata",
                *context.fixtures,
                **{"verbosity": 0, "database": db_name},
            )
        except Exception:
            # pylint: disable-next=protected-access
            context.test._rollback_atomics(context.test.cls_atomics)
            raise



"""Steps for testing authentication."""
from behave import then


@then("I should be a superuser")
def is_superuser(context):
    """Check that a user is a superuser.

    Args:
        context (behave.runner.Context): The test context.
    """
    context.test.assertTrue(getattr(context, "user", None) is not None)
    context.test.assertTrue(context.user.is_superuser)


@then("I should not be a superuser")
def is_not_superuser(context):
    """Check that a user is not a superuser.

    Args:
        context (behave.runner.Context): The test context.
    """
    context.test.assertTrue(getattr(context, "user", None) is not None)
    context.test.assertFalse(context.user.is_superuser)


@then("I should be a staff user")
def is_staff_user(context):
    """Check that a user is a staff user.

    Args:
        context (behave.runner.Context): The test context.
    """
    context.test.assertTrue(getattr(context, "user", None) is not None)
    context.test.assertTrue(context.user.is_staff)


@then("I should not be a staff user")
def is_not_staff_user(context):
    """Check that a user is not a staff user.

    Args:
        context (behave.runner.Context): The test context.
    """
    context.test.assertTrue(getattr(context, "user", None) is not None)
    context.test.assertFalse(context.user.is_staff)


@then("I should not be an active user")
def then_is_not_active_user(context):
    """Check that a user is not an active user.

    Args:
        context (behave.runner.Context): The test context.
    """
    context.test.assertTrue(getattr(context, "user", None) is not None)
    context.test.assertFalse(context.user.is_active)


@then("I should be an active user")
def then_should_be_active_user(context):
    """Check that a user is an active user.

    Args:
        context (behave.runner.Context): The test context.
    """
    context.test.assertTrue(getattr(context, "user", None) is not None)
    context.test.assertTrue(context.user.is_active)


@then('I should have a first name of "{first_name}"')
def has_first_name(context, first_name):
    """Check that a user has the correct first name.

    Args:
        context (behave.runner.Context): The test context.
        first_name (str): The first name to check.
    """
    context.test.assertTrue(getattr(context, "user", None) is not None)
    context.test.assertTrue(context.user.first_name == first_name)


@then('I should have a last name of "{last_name}"')
def has_last_name(context, last_name):
    """Check that a user has the correct last name.

    Args:
        context (behave.runner.Context): The test context.
        last_name (str): The last name to check.
    """
    context.test.assertTrue(getattr(context, "user", None) is not None)
    context.test.assertTrue(context.user.last_name == last_name)



"""Fixtures steps."""
import os
from collections import defaultdict
from pathlib import Path

from behave import when
from behave.runner import Context


@when('the data for "{apps_maybe_model}" is dumped')
def step_impl(context: Context, apps_maybe_model: str):
    """Dump the database.

    This step is used to dump the database in the context of a test.

    Args:
        context (Context): The behave context.
        apps_maybe_model (str): A comma separated list of app and
         optionally model names.
    """
    apps_and_models = (
        apps_maybe_model.split(",") if "," in apps_maybe_model else [apps_maybe_model]
    )
    context.dumped_data = defaultdict(list)

    # pylint: disable=import-outside-toplevel
    from django.conf import settings
    from django.core.management import call_command

    if not settings.FIXTURE_DIRS:
        fixtures_dir = Path(settings.BASE_DIR) / "fixtures"
    else:
        fixtures_dir = Path(settings.FIXTURE_DIRS[0])
    for app_and_model in apps_and_models:
        app, model = (
            app_and_model.split(".") if "." in app_and_model else [app_and_model, None]
        )
        if app in ["auth"]:
            if not fixtures_dir.exists():
                os.makedirs(fixtures_dir)
            output_path = (
                fixtures_dir / f"{model}.json"
                if model
                else fixtures_dir / f"{app}.json"
            )
        else:
            app_fixtures = Path(f"{app}/fixtures")
            if not app_fixtures.exists():
                os.makedirs(app_fixtures)
            output_path = (
                app_fixtures / f"{model}.json"
                if model
                else app_fixtures / f"{app}.json"
            )
        call_command(
            "dumpdata",
            *[
                f"{app}.{model}" if model else app,
                "--natural-foreign",
                "--all",
                f"--output={output_path}",
                # Maybe add support for other databases?
                # f"--database=test_default"
            ],
            format="json",
            indent=2,
            stdout=context.stdout_capture,
        )
        context.dumped_data[app].append(model)
